#include "lorentz.hh"
#include "matching.hh"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>



using namespace std;

using namespace TemplateOverlap;


///  Helper function for printing jets

void PrintEvent(const Event& event) {
  double etaJet = event.axis().eta();
  double phiJet = event.axis().eta();
   printf("%5s %15s %15s %15s %15s %15s\n","jet #", "rapidity", "phi", "pt","m","e"); // label the columns
  for (int ip = 0; ip < event.size();++ip)         
   printf("%5u %15.8f %15.8f %15.8f %15.8f %15.8f\n",ip, event[ip].eta()-etaJet,event[ip].phi()-phiJet,event[ip].pT(),event[ip].mCalc(),event[ip].e());
}


namespace TemplateOverlap{
/// The main analysis class
class MatchingMethod
{
private:

   ofstream _mystream;
   vector<jet_t> _templates; 
   double _subConeRadius;
   double _sigma;

public:
   MatchingMethod(char * outfile, Settings mySettings) : 
   _mystream(outfile), _subConeRadius(mySettings.r()), _sigma(mySettings.sigma()) {}
   void analyze(const Event jet, Event & maxTempl, double & maxOv);
   int templateRead(const string& filename);
   int templateRead(ifstream& is) ;

};

}
int main(int argc, char* argv[]) {
  
  if (argc !=4) {
    cout << "Usage:  "<< argv[0] << " <input.dat> <template.dat> <output.dat>" << endl;
    return 1;
  }
  /// Set Run Parameters 
 double subConeRadius = 0.20;
 double coneRadius = 1.0;
 double sigma = 0.333 ;
 Settings myParms(subConeRadius, coneRadius, sigma);

 /// Create the jet and best matched template containers
 Event myJet, maxTempl;
 double maxOv;

 /// Create an instance of Cone for the analysis.  
  MatchingMethod myCone(argv[3], myParms); 

 /// Load jet and templates
  myCone.templateRead(argv[2]);
  myJet.eventRead(argv[1]);
   
 /// Do the matchig and localize the best match
  myCone.analyze(myJet, maxTempl, maxOv);
   
return 0;
}
/// Search for highest matching template. 
void MatchingMethod::analyze (const Event myJet, Event & maxTempl, double & maxOv)
{ 
  /// Defining TemplateOverlap parameters
  int match_method = CONE;
  int normalize = HADRONIC;
  
   /// Source jet to matrix
  jet_t jet_mat; 
  myJet.copyTo(jet_mat);
  
  /// Create the result matrix
  vector<double>  result; 
  
  /// Do the matching and normalize
  matchTemplate(jet_mat, _templates, result, match_method, normalize);
  
  /// Localizing the best match with minMaxLoc
  double maxVal; int maxLoc; 
  maximize(result, maxVal, maxLoc );
  
  /// Result to Event 
  maxTempl.getFrom(_templates[maxLoc]);
  maxOv = maxVal; 
  
  /// Print data to an output file
  _mystream << maxVal << " " << maxLoc << std::endl;
  
  /// Print data to screen
  
  printf("-------------------------------------------------------------------------------------"); printf("\n");
  printf("-------------------------------------------------------------------------------------"); printf("\n");
  cout << "subConeRadius = " << _subConeRadius << endl;
  cout << "Max. overlap template:" << endl;
  PrintEvent(maxTempl);
  printf("-------------------------------------------------------------------------------------"); printf("\n");
  cout << " Ov"<<maxTempl.size()<<": " << maxVal << std::endl;
  printf("-------------------------------------------------------------------------------------"); printf("\n");
  
  ofstream img("locImg.dat");
  for (unsigned int i = 0; i < _templates.size();i++)
    img << _templates[i].at(0).y << " " << _templates[i].at(0).phi << " " << result.at(i) << std::endl;
  img.close();
  
  ofstream img2("maxLocImg.dat");
    img2 << _templates[maxLoc].at(0).y << " " << _templates[maxLoc].at(0).phi << " " << result.at(maxLoc) << std::endl;
    img2 << _templates[maxLoc].at(1).y << " " << _templates[maxLoc].at(1).phi << " " << result.at(maxLoc) << std::endl;

    img2.close();
 

} 


/// Load templates from file.
int MatchingMethod::templateRead(ifstream& is) {

  string line, tag;
 
 //Event particlesSave;
 
jet_t particlesSave; 
_templates.clear();

  while (getline(is,line)){
  istringstream getpro(line);
  int nupSave;
  getpro >> nupSave;
  if (!getpro) return false;

  // Reset particlesSave vector.
  particlesSave.clear(); 

  // Read in particle info one by one, and store it.
  double pup1, pup2, pup3, pup4;
  
   if (!getline(is, line)) return false;
    istringstream getall(line);
    getall  >> pup1 >> pup2 >> pup3 >> pup4;
  
  for (int ip = 0; ip < nupSave; ++ip) { 
    if (!getline(is, line)) return false;
    istringstream getall(line);
    getall  >> pup1 >> pup2 >> pup3 >> pup4;
    if (!getall) return false;   
    
   Vec4 pTemp(pup1, pup2, pup3, pup4);

   SingleParticle pAux= SingleParticle(pTemp, pTemp.pT(), pTemp.y(), pTemp.phi());
   pAux.radius = _subConeRadius;
   particlesSave.push_back(pAux);
  }
 
 if (checkTemplate (particlesSave) ) _templates.push_back(particlesSave);

		
 
  // Reading worked.
  
}
  return (_templates.size());

}
int MatchingMethod::templateRead(const string& filename) {
  ifstream file(filename.c_str());
  int n;
  n =  templateRead(file);
  file.close();
  return ( n);
  
}