#ifndef Cone_H
#define Cone_H

#include <string>
#include <vector>
#include "lorentz.hh"

using namespace std;

namespace TemplateOverlap{

enum  {CONE, GAUSSIAN};
enum  {HADRONIC, SEMILEPTONIC};

/// Parameters that define TemplateOverlap
class Settings {
private:
   double _subConeRadius;     // subCone radius
   double _coneRadius;    // characteristic jet radius
   double _sigma;  // Gaussian energy resolution relative to parton pT

public:
   Settings(const double subConeRadiusIn, const double coneRadiusIn, 
	    const double sigmaIn=0.33) :
   _subConeRadius(subConeRadiusIn), _coneRadius(coneRadiusIn), _sigma(sigmaIn) {}
   double r() const {return _subConeRadius;}
   double R0() const {return _coneRadius;}
   double sigma() const {return _sigma;}
};



/// A helper class for MatchingMethod
class SingleParticle {

public:

  // Constructors.
  SingleParticle( Vec4 pIn = 0., double pTIn = 0., double yIn = 0., 
    double phiIn = 0.) : p(pIn), pT(pTIn), y(yIn), phi(phiIn), 
    mult(1), isUsed(false), radius(0.) { }
  SingleParticle(const SingleParticle& ssj) : p(ssj.p), pT(ssj.pT),
    y(ssj.y), phi(ssj.phi), mult(ssj.mult), isUsed(ssj.isUsed), radius(ssj.radius) { }
  SingleParticle& operator=(const SingleParticle& ssj) { if (this != &ssj)
    { p = ssj.p; pT = ssj.pT; y = ssj.y; phi = ssj.phi; 
    mult = ssj.mult; isUsed = ssj.isUsed; radius =ssj.radius;} return *this; }

  // Properties of particle.
  Vec4   p;  
  double pT, y, phi;
  int    mult; 
  bool   isUsed;
  double radius; //For templates
  double deltaR2(const SingleParticle & other) const {
    double  dPhi = abs(phi - other.phi );
    if (dPhi > M_PI) dPhi = 2. * M_PI - dPhi;
    double  dEta =  y - other.y;
    return (dEta * dEta + dPhi * dPhi );
  }
  
};

typedef std::vector<SingleParticle> jet_t;
typedef std::pair<jet_t, double> templ_t;
bool checkTemplate (const jet_t & probe_template);
void matchTemplate( jet_t & jet, const  vector<jet_t> & templates, vector<double>  & result, int match_method=CONE, int normalization=HADRONIC);
void maximize(const vector<double>  & result, double & maxVal,  int & maxLoc );



/// The basic class for holding the jet constituents and the particles in a template. It also computes basic jet shapes, such as planar flow.


class Event {
 public:
  Event() {};
  bool eventRead(std::ifstream& in);
  bool eventRead(const string& tFile) ;
  Vec4& operator[](int i) {return particles[i];}
  const Vec4& operator[](int i) const {return particles[i];}
  const Vec4& axis() const {return jet_axis;}
  const Vec4& bJet(int i) const {return _bJets[i];}
  void sortByPt() {sort(particles.begin(), particles.end(), sortVecByPt);}
  int size() const {return particles.size();}
  double planarFlow() const;
  double planarFlow2(double r = 0.) const;
  double angularity(int a, double R) const;
  double thetaS() const;
  double thetaBar() const;
  double dipolarity(Vec4 j1, Vec4 j2) const;
  double dipolarity2(Vec4 j1, Vec4 j2) const;
  void setOverlap(double overlap)  {_overlap = overlap;}
  double getOverlap() const {return _overlap;}  
  void copyTo(jet_t & jet) const ;
  void getFrom(const jet_t & jet) ;
  void clear() {particles.clear();jet_axis.reset();_bJets.clear();}
  int append(Vec4 entryIn) {    
    particles.push_back(entryIn);
	jet_axis += entryIn;  
    return particles.size() - 1;
  }
  void bst(const Vec4& vec) {
    for (int i = 0; i < size(); ++i) particles[i].bst(vec);
    
  }
  
 private:
 
   vector<Vec4> particles;
   Vec4 jet_axis;
   vector<Vec4> _bJets; 
   double _overlap; 
   // Constants: could only be changed in the code itself.
   static const double TINY;

};

} //end namespace


#endif // end Event_H
